@@ -1,8 +1,6 @@
 /*
  *  version.h - Version information
  *
- *  Basilisk II (C) 1997-2000 Christian Bauer
- *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2 of the License, or
@@ -21,10 +19,29 @@
 #ifndef VERSION_H
 #define VERSION_H
 
-const int VERSION_MAJOR = 0;
-const int VERSION_MINOR = 0;
-const int VERSION_SUBMINOR = 21;
+#include "version_date.h"
+#ifdef VERSION_DATE
+#define CVS_DATE	"CVS " VERSION_DATE
+#else
+#define CVS_DATE	"CVS"
+#endif
 
-#define VERSION_STRING "ARAnyM 0.0.21"
+#define VER_MAJOR	0
+#define VER_MINOR	9
+#define VER_MICRO	9
+
+#ifndef VER_STATUS
+#define VER_STATUS
+//#define VER_STATUS	"alpha"
+//#define VER_STATUS	"beta"
+//#define VER_STATUS	"beta+" CVS_DATE
+#endif
+
+#define str(x)		_stringify (x)
+#define _stringify(x)	#x
+
+#define NAME_STRING "ARAnyM"
+
+#define VERSION_STRING	NAME_STRING " " str (VER_MAJOR) "." str (VER_MINOR) "." str (VER_MICRO) VER_STATUS 
 
 #endif
