@@ -1,20 +1,44 @@
 /*
- * $Header: /var/repos/aranym/src/ndebug.cpp,v 1.14 2002/02/01 20:19:38 milan Exp $
+ * ndebug.cpp - new ARAnyM fullscreen debugger
  *
- * MJ 2001
+ * Copyright (c) 2001-2009 Milan Jurik of ARAnyM dev team (see AUTHORS)
+ *
+ * This file is part of the ARAnyM project which builds a new and powerful
+ * TOS/FreeMiNT compatible virtual machine running on almost any hardware.
+ *
+ * ARAnyM is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * ARAnyM is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with ARAnyM; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+
 #include "sysdeps.h"
 
+#include "memory.h"
+#include "newcpu.h"
+#include "m68k.h"
+#include "cpu_emulation.h"
+#include "input.h"
+#include "cpummu.h"
+#include "fpu/fpu.h"
+
 #define DEBUG 1
 #include "debug.h"
 
-#ifndef CONFGUI
-#include "main.h"
-#include "memory.h"
-#include "newcpu.h"
-#include "cpu_emulation.h"
-#endif
+#include <cstdarg>
+#include <cstdlib>
+#include <cerrno>
+#include <assert.h>
 
 #ifndef HAVE_GNU_SOURCE
 
@@ -25,6 +49,8 @@ extern "C" int vasprintf(char **, const char *, va_list);
 
 #include "parameters.h"
 
+extern void QuitEmulator(void);
+
 #ifdef NEWDEBUG
 
 #ifndef HAVE_STRDUP
@@ -36,8 +62,8 @@ extern "C" char *strdup(const char *s)
 }
 #endif
 
-static termios savetty;
-
+termios ndebug::savetty;
+bool ndebug::issavettyvalid = false;
 unsigned int ndebug::rowlen = 78;
 //unsigned int ndebug::dbsize = 1000;
 unsigned int ndebug::warnlen = 10;
@@ -45,12 +71,21 @@ char **ndebug::dbbuffer = NULL;
 unsigned int ndebug::dbstart = 0;
 unsigned int ndebug::dbend = 0;
 unsigned int ndebug::dbfull = 0;
-unsigned int ndebug::aktualrow = 0;
+unsigned int ndebug::actualrow = 0;
 unsigned int ndebug::tp = 0;
 bool ndebug::do_skip = false;
 uaecptr ndebug::skipaddr = 0;
+bool ndebug::do_skip_value = false;
+uaecptr ndebug::value_addr = 0;
+uint32 ndebug::value;
+value_test_t  ndebug::value_test;
+
+uint32 ndebug::do_breakpoints = 0;
+bool ndebug::breakpoint[max_breakpoints];
+uaecptr ndebug::breakpoint_address[max_breakpoints];
+  
 char ndebug::old_debug_cmd[80] = "                                                                               ";
-static char *strhelp[] = {"Help:\n",
+static const char *strhelp[] = {"Help:\n",
 	" h, ?                 show this help\n",
 	" Q                    quit ARAnyM\n",
 	" R                    reset CPU\n",
@@ -59,8 +94,8 @@ static char *strhelp[] = {"Help:\n",
 	" P <number> <value>   set USP(0), ISP(1), MSP(2), VBR(3), IMASK(4), PC(5)\n",
 	" S <number> <value>   set TC(0), SRP(1), URP(2), DTT0(3), DTT1(4),\n",
 	"                      ITT0(5), ITT1(6), CACR(7), CAAR(8)\n",
-	" i                    enable IRQ\n",
-	" I                    disable IRQ\n",
+	" i                    enable/disable IRQ\n",
+	" I                    enable/disable IRQ debugging\n",
 	" g <address>          start execution at the current address or <address>\n",
 	" t <address>          step one instruction at current addres or <address>\n",
 	" z                    step through one instruction - useful for JSR, DBRA etc\n",
@@ -80,9 +115,16 @@ static char *strhelp[] = {"Help:\n",
 	" q                    change to debug type No. 0\n",
 	" j | J                browse up\n",
 	" k | K                browse down\n",
-	" r                    refresh D(bug()) | reset aktual row\n",
+	" r                    refresh D(bug()) | reset actual row\n",
+	" L <file>             save debugger's info to <file>\n",
+	" v X <addr> <value>   step forward until (<address>) = <value> (X=b,w,l)\n",
+	" V X <address>        step forward until (<address>) changed (X=b,w,l)\n",
+	" B <address>          toggle breakpoint on <address>\n",
 #ifdef FULL_HISTORY
-	" H <lines>            show history of PC",
+	" H <lines>            show history of PC\n",
+#endif
+#ifdef FULLMMU
+	" u                    dump the MMU translation tables and state\n",
 #endif
 	NULL
 };
@@ -91,33 +133,33 @@ void ndebug::ignore_ws (char **c) {
   while (**c && isspace(**c)) (*c)++;
 }
 
-void ndebug::reset_aktualrow()
+void ndebug::reset_actualrow()
 {
 	if (dbend < get_warnlen()) {
-		aktualrow = dbfull ? (dbsize - (get_warnlen() - dbend)) : 0;
+		actualrow = dbfull ? (dbsize - (get_warnlen() - dbend)) : 0;
 	} else {
-		aktualrow = dbend - get_warnlen();
+		actualrow = dbend - get_warnlen();
 	}
 }
 
-void ndebug::set_aktualrow(signed int r)
+void ndebug::set_actualrow(signed int r)
 {
 	if (r == 0)
 		return;
 	if (r < 0) {
 		r = -r;
 		r %= dbsize;
-		if (aktualrow < (unsigned int) r) {
-			aktualrow = dbfull ? (dbsize - (r - aktualrow)) : 0;
+		if (actualrow < (unsigned int) r) {
+			actualrow = dbfull ? (dbsize - (r - actualrow)) : 0;
 		} else {
-			aktualrow -= r;
+			actualrow -= r;
 		}
 	} else {
-		if ((aktualrow + r) < dbend) {
-			aktualrow += r;
+		if ((actualrow + r) < dbend) {
+			actualrow += r;
 		} else {
-			aktualrow =
-				dbfull ? (dbsize - (dbend - (aktualrow + r))) : dbend;
+			actualrow =
+				dbfull ? (dbsize - (dbend - (actualrow + r))) : dbend;
 		}
 	}
 }
@@ -125,68 +167,66 @@ void ndebug::set_aktualrow(signed int r)
 #endif
 
 #ifdef NEWDEBUG
-int ndebug::dbprintf(char *s, ...)
+int ndebug::dbprintf(const char *s, ...)
 {
 #else
-int dbprintf(char *s, ...)
+int dbprintf(const char *s, ...)
 {
 #endif
 	va_list a;
-	va_start(a, s);
 #ifdef NEWDEBUG
 	if (bx_options.startup.debugger) {
 		int i;
 		if (dbbuffer[dbend] != NULL)
 			free(dbbuffer[dbend]);
+		va_start(a, s);
 		i = vasprintf(&dbbuffer[dbend++], s, a);
+		va_end(a);
 		if (dbend == dbsize) dbend = 0;
 		if (dbstart == dbend) dbstart++;
 		if (dbstart == dbsize) dbstart = 0;
-		reset_aktualrow();
+		reset_actualrow();
 		return i;
 	} else {
 #endif
 		int i;
-		char *c;
-		i = vasprintf(&c, s, a);
-		i += fprintf(stderr, c);
+		va_start(a, s);
+		i = vfprintf(stderr, s, a);
 		i += fprintf(stderr, "\n");
-		free(c);
+		va_end(a);
 		return i;
 #ifdef NEWDEBUG
 	}
 #endif
-	va_end(a);
 }
 
 #ifdef NEWDEBUG
-int ndebug::pdbprintf(char *s, ...)
+int ndebug::pdbprintf(const char *s, ...)
 {
 #else
-int pdbprintf(char *s, ...)
+int pdbprintf(const char *s, ...)
 {
 #endif
 	int i = 0;
 	va_list a;
-	va_start(a, s);
 #ifdef NEWDEBUG
 	if (bx_options.startup.debugger) {
 		if (dbbuffer[dbend] != NULL)
 			free(dbbuffer[dbend]);
+		va_start(a, s);
 		i += vasprintf(&dbbuffer[dbend++], s, a);
+		va_end(a);
 		if (dbend == dbsize) dbend = 0;
 		if (dbstart == dbend) dbstart++;
 		if (dbstart == dbsize) dbstart = 0;
-		reset_aktualrow();
+		reset_actualrow();
 	}
 #endif
-	char *c;
-	i += vasprintf(&c, s, a);
-	i += fprintf(stderr, c);
+	va_start(a, s);
+	i = vfprintf(stderr, s, a);
 	i += fprintf(stderr, "\n");
-	free(c);
-	return i;
 	va_end(a);
+	return i;
 }
 
 
@@ -194,23 +234,18 @@ int pdbprintf(char *s, ...)
 void ndebug::warn_print(FILE * f)
 {
 	unsigned int ar;
-	char *s = (char *) malloc((rowlen + 1) * sizeof(char));
 	for (unsigned int i = 0; i < get_warnlen(); i++) {
-		ar = (aktualrow + i) % dbsize;
+		ar = (actualrow + i) % dbsize;
 		if (dbbuffer[ar] != NULL
-			&& (((dbstart <= dbend) && (ar >= dbstart) && (ar <= dbend))
-			|| (dbstart > dbend) && (i < dbsize))) {
-			snprintf(s, rowlen, dbbuffer[ar]);
-			fprintf(f, s);
-			for (unsigned int j = 0; j < (rowlen - strlen(s)); j++)
-				putc(' ', f);
+			&& ((((dbstart <= dbend) && (ar >= dbstart) && (ar <= dbend)) || (dbstart > dbend))
+				&& (i < dbsize))) {
+			fprintf(f, "%-*.*s", rowlen, rowlen, dbbuffer[ar]);
 		} else {
 			for (unsigned int j = 0; j < rowlen; j++)
 				putc(' ', f);
 		}
 		fprintf(f, "|\n");
 	}
-	free(s);
 }
 
 void ndebug::m68k_print(FILE * f)
@@ -241,11 +276,11 @@ void ndebug::m68k_print(FILE * f)
 			(unsigned long) m68k_areg(regs, 7),
 			(unsigned long) m68k_dreg(regs, 3),
 			(unsigned long) m68k_dreg(regs, 7), (unsigned long) regs.vbr);
-	fprintf(f, "T=%d%d S=%d M=%d X=%d N=%d Z=%d V=%d C=%d           TC=%04x\n",
+	fprintf(f, "T=%d%d S=%d M=%d X=%d N=%d Z=%d V=%u C=%u           TC=%04x\n",
 	    regs.t1, regs.t0, regs.s, regs.m,
-	    GET_XFLG, GET_NFLG, GET_ZFLG, GET_VFLG, GET_CFLG,
-	    		   (unsigned int) ((((int) regs.tce) << 15) |
-				(((int) regs.tcp) << 14)));
+	    (int)GET_XFLG, (int)GET_NFLG, (int)GET_ZFLG, GET_VFLG, GET_CFLG,
+	    		   (unsigned int) ((((int) regs.mmu_enabled) << 15) |
+				(((int) regs.mmu_pagesize_8k) << 14)));
 
 	fprintf(f, "CACR=%08lx DTT0=%08lx ITT0=%08lx SRP=%08lx  SFC=%d%d%d\n",
 		   (unsigned long) regs.cacr, (unsigned long) regs.dtt0,
@@ -259,14 +294,13 @@ void ndebug::m68k_print(FILE * f)
 		   (int) (regs.dfc % 2));
 
     for (unsigned int i = 0; i < 8; i++) {
-	fprintf(f, "FP%d: %g ", i, regs.fp[i]);
+	fprintf(f, "FP%d: %Lg ", i, (long double)fpu.registers[i]);
 	if (i == 3) fprintf(f, "N=%d Z=%d\n",
-	    (regs.fpsr & 0x8000000) != 0,
-	    (regs.fpsr & 0x4000000) != 0);
+	    (fpu.fpsr.condition_codes & 0x8000000) != 0,
+	    (fpu.fpsr.condition_codes & 0x4000000) != 0);
 	if (i == 7) fprintf (f, "I=%d NAN=%d\n",
-		(regs.fpsr & 0x2000000) != 0,
-		(regs.fpsr & 0x1000000) != 0);
-
+		(fpu.fpsr.condition_codes & 0x2000000) != 0,
+		(fpu.fpsr.condition_codes & 0x1000000) != 0);
     }
 }
 
@@ -297,13 +331,15 @@ void ndebug::showHelp(FILE *f)
 }
 
 void ndebug::set_Ax(char **inl) {
-	char reg;
-	if (more_params(inl)) if ((reg = readhex(inl)) < 8)
-		if (more_params(inl)) m68k_areg(regs, reg) = readhex(inl);
+	int reg;
+	if (more_params(inl))
+		if ((reg = readhex(inl)) < 8)
+			if (more_params(inl))
+				m68k_areg(regs, reg) = readhex(inl);
 }
 
 void ndebug::set_Dx(char **inl) {
-	char reg;
+	int reg;
 	if (more_params(inl)) if ((reg = readhex(inl)) < 8)
 		if (more_params(inl)) m68k_dreg(regs, reg) = readhex(inl);
 }
@@ -339,8 +375,9 @@ void ndebug::set_Sx(char **inl) {
 		if (more_params(inl)) switch (reg) {
 			case 0:
 				r = readhex(inl);
-				regs.tce = (flagtype) ((r & 0x8000) ? 1 : 0);
-				regs.tcp = (flagtype) ((r & 0x4000) ? 1 : 0);
+#ifdef FULLMMU
+				mmu_set_tc(r & 0xc000);
+#endif
 				break;
 			case 1:
 				regs.srp = readhex(inl);
@@ -369,7 +406,42 @@ void ndebug::set_Sx(char **inl) {
 		}
 }
 
-void ndebug::saveintofile(FILE *f, char **inl) {
+void ndebug::errorintofile(FILE *, char **inl) {
+	char *name;
+	FILE *fp;
+	unsigned int ar;
+	
+	if (!more_params (inl)) {
+		bug("L command needs more arguments!");
+		return;
+	}
+	
+	name = *inl;
+	while (**inl != '\0' && !isspace (**inl)) (*inl)++;
+	**inl = '\0';
+
+	fp = fopen (name, "w");
+	if (fp == NULL) {
+		bug("Couldn't open file");
+		return;
+	}
+
+        for (unsigned int i = 0; i < dbsize; i++) {
+		ar = (dbstart + i) % dbsize;
+		if (dbbuffer[ar] != NULL
+			&& ((((dbstart <= dbend) && (ar >= dbstart) && (ar <= dbend)) || (dbstart > dbend))
+				&& (i < dbsize))) {
+			fprintf(fp, "%s\n", dbbuffer[ar]);
+		}
+	}
+								
+	fclose (fp);
+	return;
+}
+
+
+
+void ndebug::saveintofile(FILE *, char **inl) {
 	uae_u32 src, len;
 	char *name;
 	FILE *fp;
@@ -386,8 +458,9 @@ void ndebug::saveintofile(FILE *f, char **inl) {
 	src = readhex (inl);
 	if (!more_params (inl)) goto S_argh;
 	len = readhex (inl);
-	if (! valid_address (src, len)) {
-		bug("Invalid memory block");
+
+	if (!ValidAtariAddr(src, false, len)) {
+		bug("Invalid memory address\n");
 		return;
 	}
 
@@ -410,7 +483,7 @@ S_argh:
 	bug("s command needs more arguments!");
 }
 
-void ndebug::loadintomemory(FILE *f, char **inl) {
+void ndebug::loadintomemory(FILE *, char **inl) {
 	uae_u32 src;
 	char *name;
 	FILE *fp;
@@ -443,7 +516,7 @@ S_argh:
 	bug("l command needs more arguments!");
 }
 
-void ndebug::writeintomem(FILE *f, char **c) {
+void ndebug::writeintomem(FILE *, char **c) {
 	uae_u32 addr = 0;
 	uae_u32 val = 0;
 	char nc;
@@ -493,43 +566,38 @@ void ndebug::writeintomem(FILE *f, char **c) {
 	}
 }
 
-void ndebug::backtrace(FILE *f, volatile unsigned int lines) {
-	volatile uaecptr st = m68k_areg(regs,7);
-	jmp_buf excep_env_old;
-	excep_env_old = excep_env;
-setjmpagain:
-        int prb = setjmp(excep_env);
-        if (prb != 0) {
-		fprintf (f, " unknown address\n");
-                goto setjmpagain;
-        }
+void ndebug::backtrace(FILE *f, unsigned int lines) {
+	uaecptr st = m68k_areg(regs,7);
 	while (lines-- > 0) {
-		fprintf (f, "%08lx: %08x\n", (unsigned long)st, get_long(st, true));
+		if (valid_address(st, 0, sz_long))
+			fprintf (f, "%08lx: %08x\n", (unsigned long)st, ReadAtariInt32(st));
+		else
+			fprintf (f, " unknown address\n");
 		st += 2;
 	}
-	excep_env = excep_env_old;
 }
 
-void ndebug::log2phys(FILE *f, uaecptr addr) {
+void ndebug::log2phys(FILE *, uaecptr addr) {
 #ifdef FULLMMU
-	if (regs.tce) {
-setjmpagain:
-		jmp_buf excep_env_old;
-		excep_env_old = excep_env;
-	        int prb = setjmp(excep_env);
-        	if (prb == 2) {
-			bug("Bus error");
-	                goto setjmpagain;
-        	} if (prb == 3) {
-			bug("Access error");
-			goto setjmpagain;
-		} if (prb != 0) {
-			bug("Unknonw access error - %d", prb);
-			goto setjmpagain;
+	if (regs.mmu_enabled) {
+		SAVE_EXCEPTION;
+		TRY(prb) {
+			bug("MMU enabled: %08lx -> %08lx", (unsigned long)addr,
+			    (unsigned long)mmu_translate(addr, regs.s, 1, 0));
 		}
-		bug("MMU enabled: %08lx -> %08lx",	(unsigned long)addr,
-			(unsigned long)mmu_decode_addr(addr, true, false));
-		excep_env = excep_env_old;
+		CATCH(prb) {
+			switch (prb) {
+			case 2:
+				bug("Bus error");
+				break;
+			case 3:
+				bug("Access error");
+				break;
+			default:
+				bug("Unknonw access error - %d", int(prb));
+			}
+		}
+		RESTORE_EXCEPTION;
 	} else {
 #endif /* FULLMMU */
 		bug("MMU disabled: %08lx", (unsigned long)addr);
@@ -571,6 +639,8 @@ int ndebug::canon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaec
 
 	show(f);
 
+	if (irqindebug) fprintf(f, "i");
+	if (ignore_irq) fprintf(f, "I");
 	fprintf(f, ">");
 	fflush(f);
 	if (fgets(input, 80, stdin) == NULL) {
@@ -598,10 +668,10 @@ int ndebug::canon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaec
 			m68k_reset();
 			break;
 		case 'i':
-			irqindebug = true;
+			irqindebug = !irqindebug;
 			break;
 		case 'I':
-			irqindebug = false;
+			ignore_irq = !ignore_irq;
 			break;
 		case 'e':
 			dump_traps (f);
@@ -626,12 +696,17 @@ int ndebug::canon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaec
 		case 'g':
 			if (more_params(&inptr)) m68k_setpc(readhex(&inptr));
 			fill_prefetch_0();
-			grabMouse(true);
-			deactivate_debugger();
+			if (wasGrabbed) grabMouse(true);
+			/*
+			 * debugger activation/deactivation need to be rewriteen
+			 * deactivate_debugger();
+			 */
+			if (do_breakpoints)
+				SPCFLAGS_SET( SPCFLAG_BRK );
 			return 0;
 		case 't':
 			if (more_params(&inptr)) m68k_setpc(readhex(&inptr));
-			set_special(SPCFLAG_BRK);
+			SPCFLAGS_SET( SPCFLAG_BRK );
 			return 0;
 		case 'd':
 			if (more_params(&inptr)) daddr = readhex(&inptr);
@@ -680,25 +755,25 @@ int ndebug::canon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaec
 				else bug("c command needs one argument!");
 			break;
 		case 'j':
-			set_aktualrow(-1);
+			set_actualrow(-1);
 			break;
 		case 'J':
-			set_aktualrow(1 - get_warnlen());
+			set_actualrow(1 - get_warnlen());
 			break;
 		case 'k':
-			set_aktualrow(1);
+			set_actualrow(1);
 			break;
 		case 'K':
-			set_aktualrow(get_warnlen() - 1);
+			set_actualrow(get_warnlen() - 1);
 			break;
 		case 'r':
-			reset_aktualrow();
+			reset_actualrow();
 			break;
 		case 'z':
 			skipaddr = nextpc;
 			do_skip = true;
 			irqindebug = true;
-			set_special(SPCFLAG_BRK);
+			SPCFLAGS_SET( SPCFLAG_BRK );
 			if (wasGrabbed) grabMouse(true);
 			return 0;
 		case 'f':
@@ -709,24 +784,150 @@ int ndebug::canon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaec
 			skipaddr = readhex(&inptr);
 			do_skip = true;
 			irqindebug = true;
-			set_special(SPCFLAG_BRK);
+			SPCFLAGS_SET( SPCFLAG_BRK );
 			if (wasGrabbed) grabMouse(true);
 			return 0;
+		case 'B':
+			{
+				unsigned int i;
+				uaecptr address;
+				if (!more_params(&inptr)) {
+					bug("f command needs one parameter!");
+					break;
+				}
+				address = readhex(&inptr);
+				for (i = 0; i < max_breakpoints; i++)
+					if ((breakpoint[i]) && (address == breakpoint_address[i]))
+						break;
+				if (i < max_breakpoints) {
+					breakpoint[i] = false;
+					do_breakpoints--;
+					bug("Breakpoint on %08x removed", address);
+					break;
+				}
+				for (i = 0; i <max_breakpoints; i++)
+					if (!breakpoint[i])
+						break;
+				if (i == max_breakpoints) {
+					bug("Maximum breakpoints reached - %d", max_breakpoints);
+					break;
+				}
+				breakpoint_address[i] = address;
+				breakpoint[i] = true;
+				do_breakpoints++;
+				bug("Breakpoint on %08x added", address);
+			}
+			break;
+		case 'v':
+			if (!more_params(&inptr)) {
+				bug("v command needs one parameter!");
+				break;
+			}
+			switch (inptr[0]) {
+				case 'b':
+					inptr += 2;
+					value_test = EQUAL_value_test_8;
+					value_addr = readhex(&inptr);
+					value = (uint8)readhex(&inptr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				case 'w':
+					inptr += 2;
+					value_test = EQUAL_value_test_16;
+					value_addr = readhex(&inptr);
+					value = (uint16)readhex(&inptr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				case 'l':
+					inptr += 2;
+					value_test = EQUAL_value_test_32;
+					value_addr = readhex(&inptr);
+					value = readhex(&inptr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				default:
+					bug("v command needs for the 1st parameter b, w or l!");
+					break;
+			}
+			break;
+		case 'V':
+			if (!more_params(&inptr)) {
+				bug("V command needs one parameter!");
+				break;
+			}
+			switch (inptr[0]) {
+				case 'b':
+					inptr += 2;
+					value_test = CHANGE_value_test_8;
+					value_addr = readhex(&inptr);
+					value = ReadAtariInt8(value_addr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				case 'w':
+					inptr += 2;
+					value_test = CHANGE_value_test_16;
+					value_addr = readhex(&inptr);
+					value = ReadAtariInt16(value_addr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				case 'l':
+					inptr += 2;
+					value_test = CHANGE_value_test_32;
+					value_addr = readhex(&inptr);
+					value = ReadAtariInt32(value_addr);
+					do_skip_value = true;
+					irqindebug = true;
+					SPCFLAGS_SET( SPCFLAG_BRK );
+					if (wasGrabbed) grabMouse(true);
+					return 0;
+				default:
+					bug("V command needs for the 1st parameter b, w or l!");
+					break;
+			}
+			break;
+		case 'L':
+			errorintofile(f, &inptr);
+			break;
 #ifdef FULL_HISTORY
 		case 'H':
 			if (!more_params(&inptr)) count = 10;
 				else count = readhex(&inptr);
 			showHistory(count);
 			break;
+#endif
+#ifdef FULLMMU
+		case 'u':
+			mmu_dump_tables();
+			break;
 #endif
 		case 'q':
 			tp = 0;
 			break;
 	}
+	if (do_breakpoints)
+		SPCFLAGS_SET( SPCFLAG_BRK );
 	return 1;
 }
 
-int ndebug::icanon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaecptr &nxmem) {
+int ndebug::icanon(FILE *f, bool, uaecptr, uaecptr &nxdis, uaecptr &nxmem) {
+	if (!issavettyvalid)
+		return(0);
+	
 	struct termios newtty;
 	char buffer[1];
 	int count;
@@ -749,7 +950,7 @@ int ndebug::icanon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uae
 		if (read(0, buffer, sizeof(buffer))) {
 			switch (buffer[0]) {
 				case 't':
-					set_special(SPCFLAG_BRK);
+					SPCFLAGS_SET( SPCFLAG_BRK );
 					fprintf(stderr, "\n");
 					tcsetattr(0, TCSAFLUSH, &savetty);
 					fflush(stderr);
@@ -762,7 +963,7 @@ int ndebug::icanon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uae
 					fflush(stderr);
 					newm68k_disasm(f, daddr, &nxdis, count);
 					tcsetattr(0, TCSAFLUSH, &newtty);
-					read(0, buffer, sizeof(buffer));
+					assert(read(0, buffer, sizeof(buffer)) >= 0);
 					break;
 				case 'm':
 					maddr = nxmem;
@@ -772,26 +973,26 @@ int ndebug::icanon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uae
 					fflush(stderr);
 					dumpmem(f, maddr, &nxmem, count);
 					tcsetattr(0, TCSAFLUSH, &newtty);
-					read(0, buffer, sizeof(buffer));
+					assert(read(0, buffer, sizeof(buffer)) >= 0);
 					break;
 				case 'T':
 					maddr = nxmem;
 					log2phys(f, maddr);
 					break;
 				case 'j':
-					set_aktualrow(-1);
+					set_actualrow(-1);
 					break;
 				case 'J':
-					set_aktualrow(1 - get_warnlen());
+					set_actualrow(1 - get_warnlen());
 					break;
 				case 'k':
-					set_aktualrow(1);
+					set_actualrow(1);
 					break;
 				case 'K':
-					set_aktualrow(get_warnlen() - 1);
+					set_actualrow(get_warnlen() - 1);
 					break;
 				case 'r':
-					reset_aktualrow();
+					reset_actualrow();
 					break;
 				case 'q':
 					tp = 0;
@@ -807,7 +1008,10 @@ int ndebug::icanon(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uae
 	return 1;
 }
 
-int ndebug::dm(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaecptr &nxmem) {
+int ndebug::dm(FILE *f, bool, uaecptr, uaecptr &, uaecptr &nxmem) {
+	if (!issavettyvalid)
+		return(0);
+
 	struct termios newtty;
 	char buffer[1];
 	int count;
@@ -834,7 +1038,7 @@ int ndebug::dm(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaecptr
 			switch (buffer[0]) {
 				case 't':
 					nxmem = maddr;
-					set_special(SPCFLAG_BRK);
+					SPCFLAGS_SET( SPCFLAG_BRK );
 					fprintf(stderr, "\n");
 					tcsetattr(0, TCSAFLUSH, &savetty);
 					fflush(stderr);
@@ -846,12 +1050,10 @@ int ndebug::dm(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaecptr
 				case 'j':
 					nxmem = maddr - 32;
 					if (nxmem > RAMSize + ROMSize + FastRAMSize) nxmem = 0;
-					if (nxmem < 0) nxmem = 0;
 					break;
 				case 'J':
 					nxmem = maddr + 32 - 32 * count;
 					if (nxmem > RAMSize + ROMSize + FastRAMSize) nxmem = 0;
-					if (nxmem < 0) nxmem = 0;
 					break;
 				case 'k':
 					nxmem = maddr + 32;
@@ -879,21 +1081,79 @@ int ndebug::dm(FILE *f, bool wasGrabbed, uaecptr nextpc, uaecptr &nxdis, uaecptr
 }
 
 void ndebug::run() {
+	if (do_skip || do_skip_value || do_breakpoints) {
+		bool success = false;
+
+		if (do_skip && (m68k_getpc() == skipaddr)) {
+			do_skip = false;
+			success = true;
+		}
+
+		if (do_skip_value) {
+			switch (value_test) {
+				case EQUAL_value_test_8:
+					if (ReadAtariInt8(value_addr) == value) {
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+				case EQUAL_value_test_16:
+					if (ReadAtariInt16(value_addr) == value){
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+				case EQUAL_value_test_32:
+					if (ReadAtariInt32(value_addr) == value){
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+				case CHANGE_value_test_8:
+					if (ReadAtariInt8(value_addr) != value) {
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+				case CHANGE_value_test_16:
+					if (ReadAtariInt16(value_addr) != value){
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+				case CHANGE_value_test_32:
+					if (ReadAtariInt32(value_addr) != value){
+						do_skip_value = false;
+						success = true;
+					}
+					break;
+			}
+		}
+
+		if (do_breakpoints) {
+			unsigned int i;
+			for (i = 0; i < max_breakpoints; i++)
+				if ((breakpoint[i]) && (breakpoint_address[i] == m68k_getpc()))
+					break;
+			if (i < max_breakpoints)
+				success = true;
+		}
+
+		if (!success) {
+			SPCFLAGS_SET( SPCFLAG_BRK );
+			return;
+		}
+	}
+
+	irqindebug = false;
+	// release keyboard and mouse control
+	bool wasGrabbed = grabMouse(false);
+
 	uaecptr nextpc, nxdis, nxmem;
 	newm68k_disasm(stderr, m68k_getpc(), &nextpc, 0);
 	nxdis = nextpc;
 	nxmem = 0;
 
-	if (do_skip && (m68k_getpc() != skipaddr)) {
-		set_special(SPCFLAG_BRK);
-		return;
-	}
-
-	irqindebug = false;
-	do_skip = false;
-	// release keyboard and mouse control
-	bool wasGrabbed = grabMouse(false);
-
 	for (;;) {
 		switch(tp) {
 			default:
@@ -918,46 +1178,49 @@ void ndebug::init()
 	}
 	for (unsigned int i = 0; i < dbsize; i++)
 		dbbuffer[i] = NULL;
+	for (unsigned int i = 0; i < max_breakpoints; i++)
+		breakpoint[i] = false;
 	if (tcgetattr(0, &savetty) == -1) {
-		fprintf(stderr, "tcgetattr error!\n");
-		exit(-1);
+		fprintf(stderr, "tcgetattr error: %d!\n", errno);
+	} else {
+		issavettyvalid = true;
 	}
 }
 
 void ndebug::nexit() {
-	tcsetattr(0, TCSAFLUSH, &savetty);
+	if (issavettyvalid)
+		tcsetattr(0, TCSAFLUSH, &savetty);
+
+	issavettyvalid = false;
 }
 
-void ndebug::dumpmem(FILE *f, volatile uaecptr addr, uaecptr * nxmem, volatile unsigned int lns)
+void ndebug::dumpmem(FILE *f, VOLATILE uaecptr addr, uaecptr * nxmem, unsigned int lns)
 {
-	jmp_buf excep_env_old;
-	excep_env_old = excep_env;
-	uaecptr a;
+	SAVE_EXCEPTION;
+	VOLATILE uaecptr a;
 	broken_in = 0;
 	for (; lns-- && !broken_in;) {
-		volatile int i;
+		VOLATILE int i;
 		fprintf(f, "%08lx: \n", (unsigned long)addr);
 		for (i = 0; i < 16; i++) {
-setjmpagain:
-        		if (setjmp(excep_env) != 0) {
+			TRY(prb) {
+				a = addr < 0xff000000 ? addr : addr & 0x00ffffff;
+				if ((addr & 0xfff00000) == 0x00f00000) {
+					fprintf(f, "%04x ", (unsigned int) HWget_w(a));
+				}
+				else {
+					fprintf(f, "%04x ", (unsigned int) ReadAtariInt16(a));
+				}
+			}
+			CATCH(prb) {
 				fprintf (f, "UA   ");
-				addr += 2;
-				if (++i == 16) break;
-		                goto setjmpagain;
 		        }
-			a = addr < 0xff000000 ? addr : addr & 0x00ffffff;
-			if ((addr & 0xfff00000) == 0x00f00000) {
-				fprintf(f, "%04x ", (unsigned int) HWget_w(a));
-			}
-			else {
-				fprintf(f, "%04x ", (unsigned int) ReadAtariInt16(a));
-			}
 			addr += 2;
 		}
 		fprintf(f, "\n");
 	}
 	*nxmem = addr;
-	excep_env = excep_env_old;
+	RESTORE_EXCEPTION;
 }
 
 uae_u32 ndebug::readhex (char v, char **c)
@@ -1086,67 +1349,35 @@ void ndebug::convertNo(char **s) {
 #ifdef FULL_HISTORY
 void ndebug::showHistory(unsigned int count) {
 	unsigned int temp = lasthist;
+#ifdef NEED_TO_DEBUG_BADLY
+	struct regstruct save_regs = regs;
+	struct flag_struct save_flags = regflags;
+#endif
 	
 	while (count-- > 0 && temp != firsthist) {
 	    if (temp == 0) temp = MAX_HIST-1; else temp--;
 	}
 	bug("History:");
 	while (temp != lasthist) {
+#ifdef NEED_TO_DEBUG_BADLY
+		regs = history[temp];
+		regflags = historyf[temp];
+		m68k_dumpstate (NULL);
+#else
 		showDisasm(history[temp]);
+#endif
 		//bug("%04x", history[temp]);
 		if (++temp == MAX_HIST) temp = 0;
 	}
+#ifdef NEED_TO_DEBUG_BADLY
+	regs = save_regs;
+	regflags = save_flags;
+#endif
 }
 #endif
 
 #endif
 
 /*
- * $Log: ndebug.cpp,v $
- * Revision 1.14  2002/02/01 20:19:38  milan
- * ioctl -> tcgetattr, tcsetattr
- *
- * Revision 1.13  2002/01/31 19:37:47  milan
- * panicbug, cleaning
- *
- * Revision 1.12  2002/01/18 23:43:25  milan
- * fgetc returns int
- * abs needs stdlib.h
- *
- * Revision 1.11  2002/01/18 20:37:47  milan
- * FixedSizeFastRAM & Makefile fixed
- *
- * Revision 1.10  2002/01/08 16:13:18  joy
- * config variables moved from global ones to bx_options struct.
- *
- * Revision 1.9  2001/12/07 00:31:31  milan
- * small bug in 'z' command of ndebug corrected, big one stays, it needs better solution
- *
- * Revision 1.8  2001/11/28 09:36:05  milan
- * cleaning
- * update of config.* file
- *
- * Revision 1.7  2001/11/21 13:29:51  milan
- * cleanning & portability
- *
- * Revision 1.6  2001/11/06 20:36:54  milan
- * MMU's corrections
- *
- * Revision 1.5  2001/10/29 08:15:45  milan
- * some changes around debuggers
- *
- * Revision 1.4  2001/10/25 22:33:18  milan
- * fullhistory's support in ndebug
- *
- * Revision 1.3  2001/10/03 11:10:03  milan
- * Px supports now P5 <addr>, PC=<addr>
- *
- * Revision 1.2  2001/10/02 19:13:28  milan
- * ndebug, malloc
- *
- * Revision 1.1  2001/07/21 18:13:29  milan
- * sclerosis, sorry, ndebug added
- *
- *
- *
- */
+vim:ts=4:sw=4:
+*/
